/////////////////
// X_PS.cpp		:	A* ·㷨ʵ
// X_PS.cpp		:	A* path searching algorithm implementing
//
// 汾0010		:	һžʮʮһ
// v0010		:	Oct.31.1997
//
// д			:	л
// written by	:	JianHua Xie
//
// 		:	Visual C++ 4.0
// Compiler		:	Visual C++ 4.0
//
// Ȩ			:	ǰ޹˾	1997
// Copyright(C)	:	WayAhead Co.Ltd.		1997
//
/////////////////
//
// ļ A* ·㷨ʵ֡
//

#include "stdafx.h"

#include "X_PS.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

////////////////////////////////////////////////////////////////////////////////

inline static int
IsLessThen(const XPS_Node& a, const XPS_Node& b)
{
	return a.GetScore() < b.GetScore();
//		|| (a.GetScore() == b.GetScore() && a.Length < b.Length)
}

////////////////////////////////////////////////////////////////////////////////

XPathSearch::XPathSearch()
{
	m_pOpenList = NULL;
	m_OpenListLimit = 0;

	for (int i = 0; i < PS_HASHBACKET_SIZE; i++)
		m_aBacket[i] = NULL;
	m_OpenListCount = 0;
	m_pNearstNode = NULL;
}

XPathSearch::~XPathSearch()
{
	Cleanup();
} 

void
XPathSearch::Flush()
{
	m_Nodes.Flush();

	for (int i = 0; i < PS_HASHBACKET_SIZE; i++) 
		m_aBacket[i] = NULL;
	m_OpenListCount = 0;
	m_pNearstNode = NULL;
}

void
XPathSearch::Cleanup()
{
	Flush();

	m_Nodes.Cleanup();
	free(m_pOpenList); m_pOpenList = NULL; m_OpenListLimit = 0;
}

int
XPathSearch::OpenList_Grow(unsigned delta)
{
	ASSERT(delta > 0);

	// calc the new limit
	unsigned newLimit = m_OpenListLimit;
	while (m_OpenListCount + delta > newLimit) {
		if (newLimit == 0)			newLimit  = 8;
		else if (newLimit >= 256)	newLimit += 16;
		else						newLimit *= 2;
	}

	if (newLimit > m_OpenListLimit) {
		if (m_OpenListLimit == 0) {
			// first allocate, use malloc()
			ASSERT(m_pOpenList == NULL);
			m_pOpenList = (XPS_Node**)malloc(newLimit * sizeof(*m_pOpenList));
			if (m_pOpenList == NULL) return FALSE;
		} else {
			// use realloc in other sense, for speedup
			ASSERT(m_pOpenList != NULL);
			XPS_Node **newOpenList = (XPS_Node**)realloc(m_pOpenList, newLimit * sizeof(*newOpenList));
			if (newOpenList == NULL) return FALSE;
			m_pOpenList = newOpenList;
		}

		m_OpenListLimit = newLimit;
	}

	return TRUE;
}

void
XPathSearch::OpenList_Pullup(register unsigned curr)
{
	register XPS_Node **pOpenList = m_pOpenList-1;

	register unsigned parent = curr >> 1;

	while (parent && IsLessThen(*pOpenList[curr], *pOpenList[parent])) {
		XPS_Node *tmp = pOpenList[curr];
		pOpenList[curr] = pOpenList[parent];
		pOpenList[parent] = tmp;
		curr = parent;
		parent >>= 1;	//parent = curr >> 1;
	}
}

void
XPathSearch::OpenList_Heapify(register unsigned curr)
{
	register XPS_Node **pOpenList = m_pOpenList-1;
	register unsigned openListCount = m_OpenListCount;

	register unsigned child = curr << 1; // curr * 2;

	while (child <= openListCount) {
		if (child < openListCount && IsLessThen(*pOpenList[child+1], *pOpenList[child]))
			child++;

		if (! IsLessThen(*pOpenList[child], *pOpenList[curr]))
			break;

		XPS_Node *tmp = pOpenList[curr];
		pOpenList[curr] = pOpenList[child];
		pOpenList[child] = tmp;
		curr = child;
		child <<= 1;	//child = child*2;
	}	
}

XPS_Node*
XPathSearch::PushNode(PS_Pos pos, XPS_Node *pParent,
					  PS_Dir dirFromParent,
					  PS_Cost actualCost, PS_Cost estimateCost,
					  PS_DirMask visitedDirs)
{
//	ASSERT(FindReachedNode(pos) == NULL);

	// we grow the open list first, for allocate fail then we can still in a safe status
	//
	if (m_OpenListCount >= m_OpenListLimit && ! OpenList_Grow(1))
		return NULL;

	// create the new node
	XPS_Node *newNode = m_Nodes.Alloc();
	if (newNode != NULL) {
		newNode->Pos = pos;
		newNode->Parent = pParent;
		newNode->ActualCost = actualCost;
		newNode->EstimateCost = estimateCost;
//		newNode->Length = pParent ? pParent->length+1 : 0;
		newNode->DirFromParent = dirFromParent;
		newNode->VisitedDirs = visitedDirs;

		// insert to the hash table
		XPS_Node **pHashEntry = &m_aBacket[HashValue(pos) % PS_HASHBACKET_SIZE];
		newNode->HashNext = *pHashEntry;
		*pHashEntry = newNode;

		// insert to the open list and update open list, make sure it is a heap
		ASSERT(m_OpenListCount < m_OpenListLimit);
		m_pOpenList[m_OpenListCount++] = newNode;
		OpenList_Pullup(m_OpenListCount);
	}
	
	// return the index of new node and update the nodes count
	return newNode;
}

inline XPS_Node*
XPathSearch::PopupNode()
{
	if (m_OpenListCount <= 0)
		return NULL;
	else {
		XPS_Node *result = m_pOpenList[0];
#ifdef _DEBUG
		for (unsigned i = 1; i < m_OpenListCount; i++) {
			ASSERT(!IsLessThen(*m_pOpenList[i], *result));
		}
#endif
		m_pOpenList[0] = m_pOpenList[--m_OpenListCount];
		OpenList_Heapify(1);
		return result;
	}
}

#ifdef _DEBUG
static int FindReachedNodeTime;
static int FindReachedNodeCount;
#endif

inline XPS_Node*
XPathSearch::FindReachedNode(const PS_Pos& pos) const
{
#ifdef _DEBUG
	FindReachedNodeTime++;
#endif
	XPS_Node *node = m_aBacket[HashValue(pos) % PS_HASHBACKET_SIZE];

	while (node != NULL) {
#ifdef _DEBUG 
		FindReachedNodeCount++;
#endif
		if (node->Pos == pos)
			break; 
		node = node->HashNext;
	}

	return node;
}

XPS_Node*
XPathSearch::FindNearstNode() const
{
	XPS_Node *pBest = NULL;

	for (unsigned i = 0; i < m_Nodes.GetCount(); i++) {
		XPS_Node *pNode = m_Nodes.Get(i);
		if (pBest == NULL || pBest->EstimateCost > pNode->EstimateCost)
			pBest = pNode;
	}

	return pBest;
}

int
XPathSearch::FindPath(PS_Pos posStart, PS_Control *pControl, unsigned cDepth)
{
	ASSERT(pControl != NULL);
	// initialize
	//
	Flush();

	// check is the start node and goal is same, then we did it.
	//
	if (pControl->IsBestPos(posStart, 0)) {
		m_pNearstNode = PushNode(posStart, NULL, PS_DIR_NONE, 0, 0, 0);
		return -1;
	}

	// add the start node to open list
	//
	PushNode(posStart, NULL, PS_DIR_NONE, 0, pControl->CalcEstimateCost(posStart), 0);

#ifdef _DEBUG
	char buf[128];
	int DupCount = 0;
	int RepCount = 0;
	int QuickSkipCount = 0;
	int OL_Count = 0;
	FindReachedNodeTime = 0;
	FindReachedNodeCount = 0;
#define DEBUG_SUMMY()	\
	(wsprintf(buf, "ALL: %d, DUP: %d, REP: %d, OL: %d, SKIP: %d\n",	\
		GetReachedNodeCount(), DupCount, RepCount, OL_Count, QuickSkipCount),	\
	OutputDebugString(buf),	\
	wsprintf(buf, "%d,%d\n", FindReachedNodeTime, FindReachedNodeCount),	\
	OutputDebugString(buf))
#else
#define DEBUG_SUMMY()	((void)0)
#endif
 
	XPS_Node *pCurrent;

	while ((pCurrent = PopupNode()) != NULL) {
		ASSERT((pCurrent->VisitedDirs & ~PS_ALL_DIR_MASK) == 0);

#ifdef __unused__
		// In a lot of sense, then happen very less.
		//
		if (pCurrent->VisitedDirs == PS_ALL_DIR_MASK) {
			QuickSkipCount++;
			continue;
		}
#endif // __unused__

		for (PS_Dir dir = 0; dir < PS_DIR_COUNT; dir++) {
			if (pCurrent->VisitedDirs & PS_DirMask(1 << dir))
				// if this direction alreay processed then skip it.
				continue;

			PS_Pos newPos;
			PS_Cost moveCost;
			PS_DirMask skipDirs;

			if (! pControl->Translate(pCurrent->Pos, dir, &newPos, &moveCost, &skipDirs))
				// if can not move to this direction then skip it,
				continue;

			ASSERT((skipDirs & ~PS_ALL_DIR_MASK) == 0);

			PS_Cost newCost = pCurrent->ActualCost + moveCost;

			// check if we reach to the goal, then the algorithm finish,
			// and return OK.
			//
			if (pControl->IsBestPos(newPos, newCost)/* && pControl->CanStayHere(newPos)*/) {
				m_pNearstNode = PushNode(newPos, pCurrent, dir, newCost, 0, skipDirs);
				if (m_pNearstNode == NULL) goto __OutOfMemory;
				DEBUG_SUMMY();
				return 1;
			}

			XPS_Node *pMoveTo = FindReachedNode(newPos);

			if (pMoveTo == NULL) { // this node not reached
				PS_Cost estimateCost = pControl->CalcEstimateCost(newPos);
				if (PushNode(newPos, pCurrent, dir, newCost, estimateCost, skipDirs) == NULL)
					goto __OutOfMemory;
			} else {
				// this node already reached
				// we need check if current path is better
				// and replace with new path
#ifdef _DEBUG
				DupCount++;
#endif
				ASSERT(pMoveTo->Parent != pCurrent);

				pMoveTo->VisitedDirs |= skipDirs;
				//??? we need check is pMoveTo->VisitedDirs == PS_ALL_DIR_MASK
				//??? and remove this node from open list

				if (newCost < pMoveTo->ActualCost) {
//					|| (newCost == pMoveTo->ActualCost && pCurrent->Length+1 < pMoveTo->Length)
#ifdef _DEBUG
					RepCount++;
#endif
					// we can't replace with new path if it isn't in open list
					// !!! OPTIMATE NEED !!!
					for (unsigned i = 0; i < m_OpenListCount; i++) {
						if (m_pOpenList[i] == pMoveTo)
							break;
					}
#ifdef _DEBUG
					if (i*2 > m_OpenListCount) OL_Count++;
#endif
					if (i < m_OpenListCount) {
						// OK, replace with new path
						//
						ASSERT(pMoveTo == m_pOpenList[i]);

						// update the cost of pMoveTo and it parent
						//
						pMoveTo->Parent = pCurrent;
						pMoveTo->ActualCost = newCost;
						pMoveTo->DirFromParent = dir;
//						pMoveTo->Length = pCurrent->Length+1;

						// update the open list, make sure it is a heap
						OpenList_Pullup(i+1);
					} else {
#ifdef _DEBUG
						QuickSkipCount++;
#endif
						// the node already remove from the open list
						// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
						ASSERT(pMoveTo->VisitedDirs == PS_ALL_DIR_MASK);
					}
				}
			} // END node no found
		} // END for

		pCurrent->VisitedDirs |= PS_ALL_DIR_MASK;

		// For not use more time, we must check some condition to stop search.
		//
		if (cDepth > 0 && GetReachedNodeCount() >= cDepth)
			break;
	} // END while

__OutOfMemory:
	m_pNearstNode = FindNearstNode();
	DEBUG_SUMMY();
	return 0;
#undef DEBUG_SUMMY
}

unsigned
XPathSearch::GetDirResult(PS_Dir *paDirs, unsigned cSize)
{
	ASSERT(m_pNearstNode != NULL);

	if (m_pNearstNode == NULL)
		return (unsigned)-1;

	const XPS_Node *pNode;
	
	pNode = m_pNearstNode;
	unsigned count = 0;
	while ((pNode = GetPrevNode(pNode)) != NULL)
		count++;
	
	if (paDirs == NULL)
		return count;

	unsigned result;

	pNode = m_pNearstNode;
	if (count > cSize) {
		do {
			pNode = GetPrevNode(pNode);
			count--;
		} while (count > cSize);
		result = 0;
	} else
		result = 1;

	while (count) {
		count--;
		paDirs[count] = pNode->DirFromParent;
		pNode = GetPrevNode(pNode);
	}

	ASSERT(pNode != NULL && GetPrevNode(pNode) == NULL);

	return result;
}

unsigned
XPathSearch::GetPosResult(PS_Pos *paPos, unsigned cSize)
{
	ASSERT(m_pNearstNode != NULL);

	if (m_pNearstNode == NULL)
		return (unsigned)-1;

	const XPS_Node *pNode;
	
	pNode = m_pNearstNode;
	unsigned count = 0;
	while ((pNode = GetPrevNode(pNode)) != NULL)
		count++;
	
	if (paPos == NULL)
		return count;

	unsigned result;

	pNode = m_pNearstNode;
	if (count > cSize) {
		do {
			pNode = GetPrevNode(pNode);
			count--;
		} while (count > cSize);
		result = 0;
	} else
		result = 1;

	while (count) {
		count--;
		paPos[count] = pNode->Pos;
		pNode = GetPrevNode(pNode);
	}

	ASSERT(pNode != NULL && GetPrevNode(pNode) == NULL);

	return result;
}

unsigned
XPathSearch::GetImmediateStatus(void *lpBuf, unsigned cSize)
{
	return 0;
}

int XPathSearch::IncreaseFind(const void *lpImmStatus, unsigned cSize,
							  PS_Control* pControl, unsigned cDepth)
{
	Flush();
	return 0;
}

void
XPathSearch::FlushResult()
{
	Flush();
}

void
XPathSearch::PackMemory()
{
	Cleanup();
}

//------------------------------------------------------------------------------

PS_Interface*
PS_New(void)
{
	return new XPathSearch();
}
